@@ -1,25 +1,22 @@
 from __future__ import annotations
 
+import asyncio
+import collections
+import contextlib
+import dataclasses
+import functools
+import logging
 import re
 import typing
-import asyncio
-import logging
-import functools
-import contextlib
-import collections
-import dataclasses
 
-import crc
-import click
-import zigpy.serial
 import async_timeout
+import click
+import crc
 import serial_asyncio
+import zigpy.serial
 
 if typing.TYPE_CHECKING:
-    try:
-        from typing import Self
-    except ImportError:
-        from typing_extensions import Self
+    from typing_extensions import Self
 
 _LOGGER = logging.getLogger(__name__)
 
@@ -60,6 +57,18 @@ def crc16_kermit(data: bytes) -> int:
     return CRC_KERMIT.checksum(data)
 
 
+def pad_to_multiple(data: bytes, multiple: int, padding: bytes) -> bytes:
+    assert len(padding) == 1
+
+    if len(data) % multiple == 0:
+        return data
+
+    num_complete_blocks = len(data) // multiple
+    padded_size = multiple * (num_complete_blocks + 1)
+
+    return data + padding * (padded_size - len(data))
+
+
 class BufferTooShort(Exception):
     """Protocol buffer requires more data to parse a packet."""
 
@@ -129,6 +138,7 @@ class SerialProtocol(asyncio.Protocol):
 
     def send_data(self, data: bytes) -> None:
         """Sends data over the connected transport."""
+        assert self._transport is not None
         data = bytes(data)
         _LOGGER.debug("Sending data %s", data)
         self._transport.write(data)
@@ -236,6 +246,7 @@ class Version:
         # 7.2.2.0 build 190
         # 4.2.2
         # SL-OPENTHREAD/2.2.2.0_GitHub-91fa1f455
+        # 4.4.0-2546d625-dirty-676fdb09
         for component in self._SEPARATORS_REGEX.split(version):
             if component.isdigit():
                 self.components.append(
@@ -256,14 +267,14 @@ class Version:
         prefix_length = min(len(our_comparable), len(their_comparable))
         return our_comparable[:prefix_length] == their_comparable[:prefix_length]
 
-    def __eq__(self, other: Self) -> bool:
-        if not isinstance(other, type(self)):
+    def __eq__(self, other: object) -> bool:
+        if not isinstance(other, Version):
             return NotImplemented
 
         return self.components == other.components
 
-    def __lt__(self, other: Self) -> bool:
-        if not isinstance(other, type(self)):
+    def __lt__(self, other: object) -> bool:
+        if not isinstance(other, Version):
             return NotImplemented
 
         our_comparable = self.comparable_components()
