@@ -1,24 +1,30 @@
 from __future__ import annotations
 
-import re
-import enum
-import typing
 import asyncio
+import enum
+import functools
+import json
 import logging
 import os.path
 import pathlib
-import functools
+import re
+import typing
 import urllib.parse
 
+import bellows.types
 import click
 import coloredlogs
-import zigpy.types
-import bellows.types
 import zigpy.ota.validators
+import zigpy.types
 
-from .gbl import GBLImage, FirmwareImageType
-from .const import DEFAULT_BAUDRATES, FW_IMAGE_TYPE_TO_APPLICATION_TYPE, ApplicationType
-from .common import CommaSeparatedNumbers, put_first, patch_pyserial_asyncio
+from .common import CommaSeparatedNumbers, patch_pyserial_asyncio, put_first
+from .const import (
+    DEFAULT_BAUDRATES,
+    FW_IMAGE_TYPE_TO_APPLICATION_TYPE,
+    ApplicationType,
+    ResetTarget,
+)
+from .firmware import FirmwareImageType, parse_firmware_image
 from .flasher import Flasher
 from .xmodemcrc import BLOCK_SIZE as XMODEM_BLOCK_SIZE, ReceiverCancelled
 
@@ -42,7 +48,7 @@ def click_enum_validator_factory(
     """Click enum validator factory."""
 
     def validator_callback(
-        ctx: click.Context, param: typing.Any, value: typing.Any
+        ctx: click.Context, param: click.Parameter, value: tuple[str]
     ) -> typing.Any:
         values = []
 
@@ -65,7 +71,7 @@ class SerialPort(click.ParamType):
 
     name = "path_or_url"
 
-    def convert(self, value, param, ctx):
+    def convert(self, value: tuple | str, param: click.Parameter, ctx: click.Context):
         if isinstance(value, tuple):
             return value
 
@@ -100,7 +106,7 @@ class SerialPort(click.ParamType):
 
 @click.group()
 @click.option("-v", "--verbose", count=True)
-@click.option("--device", type=SerialPort(), required=True)
+@click.option("--device", type=SerialPort())
 @click.option("--baudrate", hidden=True)
 @click.option(
     "--bootloader-baudrate",
@@ -133,19 +139,32 @@ class SerialPort(click.ParamType):
     callback=click_enum_validator_factory(ApplicationType),
     show_default=True,
 )
+@click.option(
+    "--bootloader-reset",
+    type=click.Choice([t.value for t in ResetTarget]),
+)
 @click.pass_context
 def main(
-    ctx,
-    verbose,
-    device,
-    baudrate,
-    bootloader_baudrate,
-    cpc_baudrate,
-    ezsp_baudrate,
-    spinel_baudrate,
-    probe_method,
-):
-    coloredlogs.install(level=LOG_LEVELS[min(len(LOG_LEVELS) - 1, verbose)])
+    ctx: click.Context,
+    verbose: bool,
+    device: str,
+    baudrate: int | None,
+    bootloader_baudrate: list[int],
+    cpc_baudrate: list[int],
+    ezsp_baudrate: list[int],
+    spinel_baudrate: list[int],
+    probe_method: list[ApplicationType],
+    bootloader_reset: str | None,
+) -> None:
+    coloredlogs.install(
+        fmt=(
+            "%(asctime)s.%(msecs)03d"
+            " %(hostname)s"
+            " %(name)s"
+            " %(levelname)s %(message)s"
+        ),
+        level=LOG_LEVELS[min(len(LOG_LEVELS) - 1, verbose)],
+    )
 
     # Override all application baudrates if a specific value is provided
     if ctx.get_parameter_source("baudrate") != click.core.ParameterSource.DEFAULT:
@@ -155,6 +174,17 @@ def main(
             " (see `--help`)"
         )
 
+    # To maintain some backwards compatibility, make `--device` required only when we
+    # are actually invoking a command that interacts with a device
+    if ctx.get_parameter_source(
+        "device"
+    ) == click.core.ParameterSource.DEFAULT and ctx.invoked_subcommand not in (
+        dump_gbl_metadata.name
+    ):
+        # Replicate the "Error: Missing option" traceback
+        param = next(p for p in ctx.command.params if p.name == "device")
+        raise click.MissingParameter(ctx=ctx, param=param)
+
     ctx.obj = {
         "verbosity": verbose,
         "flasher": Flasher(
@@ -166,15 +196,62 @@ def main(
                 ApplicationType.SPINEL: spinel_baudrate,
             },
             probe_methods=probe_method,
+            bootloader_reset=bootloader_reset,
         ),
     }
 
 
+@main.command()
+@click.pass_context
+@click.option("--firmware", type=click.File("rb"), required=True, show_default=True)
+@click_coroutine
+async def dump_gbl_metadata(ctx: click.Context, firmware: typing.BinaryIO) -> None:
+    # Parse and validate the firmware image
+    firmware_data = firmware.read()
+    firmware.close()
+
+    try:
+        fw_image = parse_firmware_image(firmware_data)
+    except zigpy.ota.validators.ValidationError as e:
+        raise click.ClickException(
+            f"{firmware.name!r} does not appear to be a valid firmware image: {e!r}"
+        )
+
+    try:
+        metadata = fw_image.get_nabucasa_metadata()
+    except KeyError:
+        metadata_obj = None
+    else:
+        metadata_obj = metadata.original_json
+        _LOGGER.info("Extracted firmware metadata: %s", metadata)
+
+    print(json.dumps(metadata_obj))
+
+
+@main.command()
+@click.pass_context
+@click_coroutine
+async def probe(ctx: click.Context) -> None:
+    flasher = ctx.obj["flasher"]
+
+    try:
+        await flasher.probe_app_type()
+    except RuntimeError as e:
+        raise click.ClickException(str(e)) from e
+
+    if flasher.app_type == ApplicationType.EZSP:
+        _LOGGER.info("Dumping EmberZNet Config")
+        try:
+            await flasher.dump_emberznet_config()
+        except RuntimeError as e:
+            raise click.ClickException(str(e)) from e
+
+
 @main.command()
 @click.pass_context
 @click.option("--ieee", required=True, type=zigpy.types.EUI64.convert)
 @click_coroutine
-async def write_ieee(ctx, ieee):
+async def write_ieee(ctx: click.Context, ieee: zigpy.types.EUI64) -> None:
     new_eui64 = bellows.types.EmberEUI64(ieee)
 
     try:
@@ -190,17 +267,19 @@ async def write_ieee(ctx, ieee):
 @click.option("--allow-downgrades", is_flag=True, default=False, show_default=True)
 @click.option("--allow-cross-flashing", is_flag=True, default=False, show_default=True)
 @click.option("--yellow-gpio-reset", is_flag=True, default=False, show_default=True)
+@click.option("--sonoff-reset", is_flag=True, default=False, show_default=True)
 @click.pass_context
 @click_coroutine
 async def flash(
-    ctx,
-    firmware,
-    force,
-    ensure_exact_version,
-    allow_downgrades,
-    allow_cross_flashing,
-    yellow_gpio_reset,
-):
+    ctx: click.Context,
+    firmware: typing.BinaryIO,
+    force: bool,
+    ensure_exact_version: bool,
+    allow_downgrades: bool,
+    allow_cross_flashing: bool,
+    yellow_gpio_reset: bool,
+    sonoff_reset: bool,
+) -> None:
     flasher = ctx.obj["flasher"]
 
     # Parse and validate the firmware image
@@ -208,14 +287,14 @@ async def flash(
     firmware.close()
 
     try:
-        gbl_image = GBLImage.from_bytes(firmware_data)
-    except zigpy.ota.validators.ValidationError as e:
+        fw_image = parse_firmware_image(firmware_data)
+    except (zigpy.ota.validators.ValidationError, ValueError) as e:
         raise click.ClickException(
-            f"{firmware.name!r} does not appear to be a valid GBL image: {e!r}"
+            f"{firmware.name!r} does not appear to be a valid firmware image: {e!r}"
         )
 
     try:
-        metadata = gbl_image.get_nabucasa_metadata()
+        metadata = fw_image.get_nabucasa_metadata()
     except KeyError:
         metadata = None
     else:
@@ -239,16 +318,27 @@ async def flash(
         # Probe with the firmware's baudrate first
         if (
             metadata.baudrate is not None
-            and ctx.parent.get_parameter_source(app_type.name)
+            and ctx.parent.get_parameter_source(f"{app_type.value}_baudrate")
             == click.core.ParameterSource.DEFAULT
         ):
-            _LOGGER.debug("Probing with %s baudrate first", app_type)
+            _LOGGER.debug("Probing with %s baudrate first", metadata.baudrate)
             flasher._baudrates[app_type] = put_first(
                 flasher._baudrates[app_type], [metadata.baudrate]
             )
+    # Maintain backward compatibility with the deprecated reset flags
+    reset_msg = (
+        "The '%s' flag is deprecated. Use '--bootloader-reset' "
+        "instead, see --help for details."
+    )
+    if yellow_gpio_reset:
+        flasher._reset_target = ResetTarget.YELLOW
+        _LOGGER.info(reset_msg, "--yellow-gpio-reset")
+    elif sonoff_reset:
+        flasher._reset_target = ResetTarget.SONOFF
+        _LOGGER.info(reset_msg, "--sonoff-reset")
 
     try:
-        await flasher.probe_app_type(yellow_gpio_reset=yellow_gpio_reset)
+        await flasher.probe_app_type()
     except RuntimeError as e:
         raise click.ClickException(str(e)) from e
 
@@ -292,17 +382,17 @@ async def flash(
                     flasher.app_baudrate,
                     metadata.baudrate,
                 )
-            elif app_version.compatible_with(fw_version):
-                _LOGGER.info(
-                    "Firmware version %s is flashed, not re-installing", app_version
-                )
-                return
-            elif ensure_exact_version and not app_version.compatible_with(fw_version):
+            elif ensure_exact_version and app_version != fw_version:
                 _LOGGER.info(
                     "Firmware version %s does not match expected version %s",
                     fw_version,
                     app_version,
                 )
+            elif app_version.compatible_with(fw_version):
+                _LOGGER.info(
+                    "Firmware version %s is flashed, not re-installing", app_version
+                )
+                return
             elif not allow_downgrades and app_version > fw_version:
                 _LOGGER.info(
                     "Firmware version %s does not upgrade current version %s",
@@ -331,7 +421,7 @@ async def flash(
     with pbar:
         try:
             await flasher.flash_firmware(
-                gbl_image,
+                fw_image,
                 run_firmware=True,
                 progress_callback=lambda current, _: pbar.update(XMODEM_BLOCK_SIZE),
             )
