@@ -1,65 +1,34 @@
 from __future__ import annotations
 
-import time
-import typing
 import asyncio
-import logging
 import dataclasses
+import logging
+import typing
 
-import bellows.ezsp
 import async_timeout
-import bellows.types
 import bellows.config
+import bellows.ezsp
+import bellows.types
 
+from .common import (
+    PROBE_TIMEOUT,
+    SerialProtocol,
+    Version,
+    connect_protocol,
+    pad_to_multiple,
+)
+from .const import DEFAULT_BAUDRATES, GPIO_CONFIGS, ApplicationType, ResetTarget
 from .cpc import CPCProtocol
-from .gbl import GBLImage
-from .const import DEFAULT_BAUDRATES, ApplicationType
-from .common import PROBE_TIMEOUT, Version, connect_protocol
-from .spinel import SpinelProtocol
 from .emberznet import connect_ezsp
+from .firmware import FirmwareImage
+from .gecko_bootloader import GeckoBootloaderProtocol, NoFirmwareError
+from .gpio import find_gpiochip_by_label, send_gpio_pattern
+from .spinel import SpinelProtocol
 from .xmodemcrc import BLOCK_SIZE as XMODEM_BLOCK_SIZE
-from .gecko_bootloader import NoFirmwareError, GeckoBootloaderProtocol
 
 _LOGGER = logging.getLogger(__name__)
 
-
-def _send_gpio_pattern(pin_states: dict[int, list[bool]], toggle_delay: float) -> None:
-    # `gpiod` isn't available on Windows
-    import gpiod
-
-    chip = gpiod.chip(0, gpiod.chip.OPEN_BY_NUMBER)
-    lines = {pin: chip.get_line(pin) for pin in pin_states.keys()}
-
-    config = gpiod.line_request()
-    config.consumer = "universal-silabs-flasher"
-    config.request_type = gpiod.line_request.DIRECTION_OUTPUT
-
-    try:
-        # Open the pins and set their initial states
-        for pin, line in lines.items():
-            state = pin_states[pin][0]
-            line.request(config, int(state))
-
-        time.sleep(toggle_delay)
-
-        # Send all subsequent states
-        for i in range(1, len(pin_states[pin])):
-            for pin, line in lines.items():
-                line.set_value(int(pin_states[pin][i]))
-
-            time.sleep(toggle_delay)
-    finally:
-        # Clean up and ensure the GPIO pins are reset to inputs
-        for line in lines.values():
-            line.set_direction_input()
-
-
-async def send_gpio_pattern(
-    pin_states: dict[int, list[bool]], toggle_delay: float
-) -> None:
-    await asyncio.get_running_loop().run_in_executor(
-        None, _send_gpio_pattern, pin_states, toggle_delay
-    )
+EZSP_BOOTLOADER_LAUNCH_DELAY = 5
 
 
 @dataclasses.dataclass(frozen=True)
@@ -81,6 +50,7 @@ class Flasher:
             ApplicationType.SPINEL,
         ),
         device: str,
+        bootloader_reset: str | None = None,
     ):
         self._baudrates = baudrates
         self._probe_methods = probe_methods
@@ -91,17 +61,38 @@ class Flasher:
         self.app_baudrate: int | None = None
         self.bootloader_baudrate: int | None = None
 
-    async def enter_yellow_bootloader(self):
-        _LOGGER.info("Triggering Yellow bootloader")
-
-        await send_gpio_pattern(
-            pin_states={
-                24: [True, False, False],
-                25: [True, False, True],
-            },
-            toggle_delay=0.1,
+        self._reset_target: ResetTarget | None = (
+            ResetTarget(bootloader_reset) if bootloader_reset else None
         )
 
+    async def enter_bootloader_reset(self, target):
+        _LOGGER.info(f"Triggering {target.value} bootloader")
+        if target in GPIO_CONFIGS.keys():
+            config = GPIO_CONFIGS[target]
+            if "chip" not in config.keys():
+                _LOGGER.warning(
+                    f"When using {target.value} bootloader reset "
+                    + "ensure no other CP2102 USB serial devices are connected."
+                )
+                config["chip"] = await find_gpiochip_by_label(config["chip_name"])
+            await send_gpio_pattern(
+                config["chip"], config["pin_states"], config["toggle_delay"]
+            )
+        else:
+            await self.enter_serial_bootloader()
+
+    async def enter_serial_bootloader(self):
+        baudrate = self._baudrates[ApplicationType.GECKO_BOOTLOADER][0]
+        async with connect_protocol(self._device, baudrate, SerialProtocol) as sonoff:
+            serial = sonoff._transport.serial
+            serial.dtr = False
+            serial.rts = True
+            await asyncio.sleep(0.1)
+            serial.dtr = True
+            serial.rts = False
+            await asyncio.sleep(0.5)
+            serial.dtr = False
+
     def _connect_gecko_bootloader(self, baudrate: int):
         return connect_protocol(self._device, baudrate, GeckoBootloaderProtocol)
 
@@ -123,6 +114,9 @@ class Flasher:
 
                 if run_firmware:
                     await gecko.run_firmware()
+                    _LOGGER.info("Launched application from bootloader")
+
+            await asyncio.sleep(1)
         except NoFirmwareError:
             _LOGGER.warning("No application can be launched")
             return ProbeResult(
@@ -170,23 +164,24 @@ class Flasher:
     async def probe_app_type(
         self,
         types: typing.Iterable[ApplicationType] | None = None,
-        *,
-        yellow_gpio_reset: bool = False,
     ) -> None:
         if types is None:
             types = self._probe_methods
 
-        if yellow_gpio_reset:
-            await self.enter_yellow_bootloader()
+        # Reset into bootloader
+        if self._reset_target:
+            await self.enter_bootloader_reset(self._reset_target)
 
         bootloader_probe = None
 
-        # Only run firmware from the bootloader if we have other probe methods
+        # Only run firmware from the bootloader if we have bootloader reset and
+        # other probe methods
         only_probe_bootloader = types == [ApplicationType.GECKO_BOOTLOADER]
+        run_firmware = self._reset_target and not only_probe_bootloader
         probe_funcs = {
             ApplicationType.GECKO_BOOTLOADER: (
                 lambda baudrate: self.probe_gecko_bootloader(
-                    run_firmware=(not only_probe_bootloader), baudrate=baudrate
+                    run_firmware=run_firmware, baudrate=baudrate
                 )
             ),
             ApplicationType.CPC: self.probe_cpc,
@@ -218,13 +213,6 @@ class Flasher:
                 bootloader_probe = result
                 self.bootloader_baudrate = bootloader_probe.baudrate
 
-                # We cannot assume that the bootloader is the only running application
-                # if we did not try to start an application
-                if only_probe_bootloader:
-                    break
-
-                _LOGGER.info("Launched application from bootloader, continuing probing")
-
             if result.continue_probing:
                 continue
 
@@ -233,22 +221,18 @@ class Flasher:
             self.app_baudrate = result.baudrate
             break
         else:
-            if bootloader_probe is None:
+            if bootloader_probe and self._reset_target:
+                # We have no valid application image but can still re-enter the
+                # bootloader
+                await self.enter_bootloader_reset(self._reset_target)
+
+                self.app_type = ApplicationType.GECKO_BOOTLOADER
+                self.app_version = bootloader_probe.version
+                self.app_baudrate = bootloader_probe.baudrate
+                self.bootloader_baudrate = bootloader_probe.baudrate
+                _LOGGER.warning("Bootloader did not launch a valid application")
+            else:
                 raise RuntimeError("Failed to probe running application type")
-            elif not yellow_gpio_reset and only_probe_bootloader:
-                raise RuntimeError(
-                    "Cannot reboot back into bootloader from unknown application"
-                )
-
-            # We have no valid application image but can still enter the bootloader
-            if yellow_gpio_reset:
-                await self.enter_yellow_bootloader()
-
-            self.app_type = ApplicationType.GECKO_BOOTLOADER
-            self.app_version = bootloader_probe.version
-            self.app_baudrate = bootloader_probe.baudrate
-            self.bootloader_baudrate = bootloader_probe.baudrate
-            _LOGGER.warning("Bootloader did not launch a valid application")
 
         _LOGGER.info(
             "Detected %s, version %s at %s baudrate (bootloader baudrate %s)",
@@ -275,12 +259,20 @@ class Flasher:
                     await spinel.enter_bootloader()
         elif self.app_type is ApplicationType.EZSP:
             async with self._connect_ezsp(self.app_baudrate) as ezsp:
-                res = await ezsp.launchStandaloneBootloader(0x01)
-
-                if res[0] != bellows.types.EmberStatus.SUCCESS:
-                    raise RuntimeError(
-                        f"EmberZNet could not enter the bootloader: {res[0]!r}"
+                try:
+                    res = await ezsp.launchStandaloneBootloader(0x01)
+                except asyncio.TimeoutError:
+                    _LOGGER.warning(
+                        "Application failed to respond to bootloader launching command."
+                        " Assuming bootloader has launched."
                     )
+                else:
+                    if res[0] != bellows.types.EmberStatus.SUCCESS:
+                        raise RuntimeError(
+                            f"EmberZNet could not enter the bootloader: {res[0]!r}"
+                        )
+
+                    await asyncio.sleep(EZSP_BOOTLOADER_LAUNCH_DELAY)
         else:
             raise RuntimeError(f"Invalid application type: {self.app_type}")
 
@@ -290,17 +282,14 @@ class Flasher:
 
     async def flash_firmware(
         self,
-        firmware: GBLImage,
+        firmware: FirmwareImage,
         run_firmware: bool = True,
         progress_callback: typing.Callable[[int, int], typing.Any] | None = None,
     ) -> None:
         data = firmware.serialize()
 
         # Pad the image to the XMODEM block size
-        if len(data) % XMODEM_BLOCK_SIZE != 0:
-            num_complete_blocks = len(data) // XMODEM_BLOCK_SIZE
-            padded_size = XMODEM_BLOCK_SIZE * (num_complete_blocks + 1)
-            data += b"\xFF" * (padded_size - len(data))
+        data = pad_to_multiple(data, XMODEM_BLOCK_SIZE, b"\xFF")
 
         async with self._connect_gecko_bootloader(self.bootloader_baudrate) as gecko:
             await gecko.probe()
@@ -309,6 +298,17 @@ class Flasher:
             if run_firmware:
                 await gecko.run_firmware()
 
+    async def dump_emberznet_config(self) -> None:
+        if self.app_type != ApplicationType.EZSP:
+            raise RuntimeError(f"Device is not running EmberZNet: {self.app_type}")
+
+        async with self._connect_ezsp(self.app_baudrate) as ezsp:
+            for config in ezsp.types.EzspConfigId:
+                v = await ezsp.getConfigurationValue(config)
+                if v[0] == bellows.types.EzspStatus.ERROR_INVALID_ID:
+                    continue
+                print(f"{config.name}={v[1]}")
+
     async def write_emberznet_eui64(self, new_eui64: bellows.types.EUI64) -> bool:
         await self.probe_app_type()
 
